#include "SplitFlap.h"
#include <AccelStepper.h> //https://github.com/waspinator/AccelStepper, reference: http://www.airspayce.com/mikem/arduino/AccelStepper/classAccelStepper.html
#include <Arduino.h>
#include <Wire.h>

// Setting constants to specific numbers
const byte CALIBRATE = 1;
const byte GO_TO_LETTER = 2;
const byte SPIN = 3;
const byte TEST = 4;

const int FollowerAddress = 11;
const int hallPin = 2;
volatile bool startFound = false;

// Array storing each symbol on the split flap. This should mirror the characters shown on your final split flap.
char flap[] = {'<', '.', '?', '!', '`', '-', '(', ')', '1', '2', '3', '4', '5', '6', '7', '8', '9', '0', '^', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '>', '#', '@', '$', '%'};

// These refer to the number of steps needed to rotate one flap and to rotate 360 degrees across the whole split flap.
// stepsPerIndex is calculated by checking stepsPerRevolution / (# of flaps in split flaps)
const int stepsPerIndex = 8; 
const int stepsPerRevolution = 200;

// How quickly the steppermotor should spin. This might change if you are using a different stepper motor. 
// Faster is not necessarily recommended as it might cause precision issues in getting to the correct flap.
const int speedOfSpool = 100;

// Dynamic variable to store current flap the system is on.
int currFlap = 0;

// Flag storing whether or not the follower's current command (calibrating, moving to letter, etc.) is still being executed or completed.
bool completedCurrentState = false;
bool ledTest = LOW;

// Individual offset which can be adjusted if different follower split flaps are offset. Defaults to 0, should not be needed unless something went wrong in the hardware.
const int offset = 0;

// Points to the physical stepper motor. We will send commands to the stepper motor by calling module0motor.
AccelStepper module0motor = AccelStepper(AccelStepper::DRIVER, 4, 3); //step pin, direction pin

// Points to an instance of the SplitFlap class defined in SplitFlap.h.
SplitFlap module0 = SplitFlap( //a split flap needs a pointer to an accelStepper and pins and info
    &module0motor, //https://www.airspayce.com/mikem/arduino/AccelStepper/classAccelStepper.html#a3bc75bd6571b98a6177838ca81ac39ab
    2, stepsPerRevolution, speedOfSpool, 100, 50, offset); //zero sensor pin, stepsPerRev, speed, accel, numberOfFlaps, zeroPositionOffset
    
SF_MAKE_ISR_MACRO(module0); //create interrupt service routine

// enum for Follower States
enum FollowerState {
  Waiting,
  Calibrating,
  SearchingForLetter,
  Spinning,
};

// Variable storing current state of this follower
FollowerState state;

void setup()
{
    // Sets up communication with leader arduino
    module0.setUpInterrupts(module0_SF_ISR);
    module0.begin();
    Wire.begin(FollowerAddress);
    Wire.onRequest(requestEvent); 
    Wire.onReceive(receiveEvent); 
    
    pinMode(LED_BUILTIN, OUTPUT);

    Serial.begin(9600);
    Serial.println("follower");

    currFlap = 0;
    completedCurrentState = false;
    state = Waiting;
}

// Repeats on loop while program is running
void loop() {

  // Switch statement which determines what this follower should be doing based on current state and updates state.
  switch (state) {

    // When calibrating, we run until calibration has completed.
    case (Calibrating):
      if(completedCurrentState == false && !module0.motor->isRunning()) {
        completedCurrentState = true;
        state = Waiting;
      }
      module0.run();
      break;

    // We run until arrived at target letter.
    case (SearchingForLetter):
      if(completedCurrentState == false && !module0.motor->isRunning()) {
        completedCurrentState = true;
        state = Waiting;
      }
      module0.run();
      break;

    // During testing, runs the motor
    case (Spinning):
      module0.runSpeed();
      break;
      
    default:
      module0.run();
      break;
  }
  
}

// Handles communication with leader. Sends 1 or 0 depending on whether current state is completed or currently being executed.
void requestEvent() {
    if(completedCurrentState){
        Wire.write(1);
    } else {
        Wire.write(0);
    }
}

// Handler which runs automatically when a command/message is received from the leader arduino.
// Runs everytime a command is received.
void receiveEvent(int numBytes){
  
    Serial.println("received event");
    completedCurrentState = false; // updates current state to false to show that a command is now being executed.

    // stores local values
    int command;
    char value;
    
    while (1 < Wire.available()){
      command = Wire.read();
    }
    
    value = Wire.read();
    Serial.println(value);

    // Switch state executing different commands based on which command was received.
    switch(command) {
      
      case CALIBRATE:
          calibrate();
          break;
          
      case GO_TO_LETTER:
          goToLetter(value);
          break;
          
      case SPIN:
          toggleSpin();
          
      case TEST:
          test();
          break;
          
      default:
        break;
    }
    
}

// Runs whenever calibrate command is received. Updates state for calibration to run in loop.
void calibrate() { 
  Serial.println("calibrating");
  module0.display(0);
  state = Calibrating;
}

// Runs to specific target letter.
void goToLetter(char target) {
  
    int targetFlap = 0; // defines index of target flap

    // identifies targetFlap index through flap array
    for (int i = 0; i < 50; i++) {
        if (flap[i] == target) {
          targetFlap = i; 
        }
    }
    
    module0.display(targetFlap); 
    currFlap = targetFlap; // updates current flap variable
    state = SearchingForLetter; // updates state
}

// Spins split flap constantly for debugging purposes. Updates state.
void toggleSpin() {
  if(state == Spinning) {
    state = Waiting;
  }
  else {
    module0.setSpeed(speedOfSpool);
    state = Spinning;
  }
}

// Test routine command. Not for intended final use.
void test() {
  Serial.println("test");
  digitalWrite(LED_BUILTIN, ledTest);
  ledTest = !ledTest;
  completedCurrentState = true;
  Serial.print("completed test");
}
